from dataclasses import dataclass, field

from ..configDict import itemType
from ..inventory.item import InventoryItem
from ..location import weaponslot


@dataclass(frozen=True)
class Keyblade(InventoryItem):
    id: int
    name: str
    weaponslot_id: int
    type: itemType = field(init=False, default=itemType.KEYBLADE)


Oathkeeper = Keyblade(42, "Oathkeeper", weaponslot_id=weaponslot.LocationId.Oathkeeper)
Oblivion = Keyblade(43, "Oblivion", weaponslot_id=weaponslot.LocationId.Oblivion)
Pureblood = Keyblade(
    71, "Pureblood", weaponslot_id=weaponslot.LocationId.Pureblood
)  # Anti-Form Dummy
StarSeeker = Keyblade(
    480, "Star Seeker", weaponslot_id=weaponslot.LocationId.StarSeeker
)
HiddenDragon = Keyblade(
    481, "Hidden Dragon", weaponslot_id=weaponslot.LocationId.HiddenDragon
)
HerosCrest = Keyblade(
    484, "Hero's Crest", weaponslot_id=weaponslot.LocationId.HerosCrest
)
Monochrome = Keyblade(485, "Monochrome", weaponslot_id=weaponslot.LocationId.Monochrome)
FollowTheWind = Keyblade(
    486, "Follow the Wind", weaponslot_id=weaponslot.LocationId.FollowTheWind
)
CircleOfLife = Keyblade(
    487, "Circle of Life", weaponslot_id=weaponslot.LocationId.CircleOfLife
)
PhotonDebugger = Keyblade(
    488, "Photon Debugger", weaponslot_id=weaponslot.LocationId.PhotonDebugger
)
GullWing = Keyblade(489, "Gull Wing", weaponslot_id=weaponslot.LocationId.GullWing)
RumblingRose = Keyblade(
    490, "Rumbling Rose", weaponslot_id=weaponslot.LocationId.RumblingRose
)
GuardianSoul = Keyblade(
    491, "Guardian Soul", weaponslot_id=weaponslot.LocationId.GuardianSoul
)
WishingLamp = Keyblade(
    492, "Wishing Lamp", weaponslot_id=weaponslot.LocationId.WishingLamp
)
DecisivePumpkin = Keyblade(
    493, "Decisive Pumpkin", weaponslot_id=weaponslot.LocationId.DecisivePumpkin
)
SleepingLion = Keyblade(
    494, "Sleeping Lion", weaponslot_id=weaponslot.LocationId.SleepingLion
)
SweetMemories = Keyblade(
    495, "Sweet Memories", weaponslot_id=weaponslot.LocationId.SweetMemories
)
MysteriousAbyss = Keyblade(
    496, "Mysterious Abyss", weaponslot_id=weaponslot.LocationId.MysteriousAbyss
)
FatalCrest = Keyblade(
    497, "Fatal Crest", weaponslot_id=weaponslot.LocationId.FatalCrest
)
BondOfFlame = Keyblade(
    498, "Bond of Flame", weaponslot_id=weaponslot.LocationId.BondOfFlame
)
Fenrir = Keyblade(499, "Fenrir", weaponslot_id=weaponslot.LocationId.Fenrir)
UltimaWeapon = Keyblade(
    500, "Ultima Weapon", weaponslot_id=weaponslot.LocationId.UltimaWeapon
)
TwoBecomeOne = Keyblade(
    543, "Two Become One", weaponslot_id=weaponslot.LocationId.TwoBecomeOne
)
WinnersProof = Keyblade(
    544, "Winner's Proof", weaponslot_id=weaponslot.LocationId.WinnersProof
)


def get_all_keyblades() -> list[Keyblade]:
    return get_locking_keyblades() + [
        Oblivion,
        Pureblood,
        StarSeeker,
        GullWing,
        GuardianSoul,
        MysteriousAbyss,
        FatalCrest,
        Fenrir,
        UltimaWeapon,
    ]


def get_locking_keyblade_names() -> list[str]:
    return [i.name for i in get_locking_keyblades()]


def get_locking_keyblades() -> list[Keyblade]:
    return [
        Oathkeeper,
        HiddenDragon,
        HerosCrest,
        Monochrome,
        FollowTheWind,
        CircleOfLife,
        PhotonDebugger,
        WishingLamp,
        DecisivePumpkin,
        SleepingLion,
        SweetMemories,
        BondOfFlame,
        TwoBecomeOne,
        WinnersProof,
        RumblingRose,
    ]



from dataclasses import dataclass
from enum import Enum


class ObjectiveType(str, Enum):
    BOSS = "Boss"
    WORLDPROGRESS = "WorldProgress"
    FIGHT = "Fight"


class ObjectiveDifficulty(str, Enum):
    EARLY = "Early"
    MIDDLE = "Middle"
    LATE = "Late"
    LATEST = "Latest"


@dataclass(frozen=True)
class KH2Objective:
    Name: str
    Location: Enum
    Type: ObjectiveType
    Difficulty: ObjectiveDifficulty = ObjectiveDifficulty.EARLY


def get_last_story_objectives():
    from .location import (
        simulatedtwilighttown,
        twilighttown,
        hollowbastion,
        landofdragons,
        beastscastle,
        disneycastle,
        portroyal,
        agrabah,
        halloweentown,
        pridelands,
        spaceparanoids,
        worldthatneverwas,
        olympuscoliseum,
        hundredacrewood,
        atlantica,
        formlevel,
        puzzlereward,
    )

    return [
        KH2Objective(
            "Defeat Axel II",
            simulatedtwilighttown.CheckLocation.Axel2,
            ObjectiveType.BOSS,
        ),
        KH2Objective(
            "Fight alongside Axel",
            twilighttown.CheckLocation.BetwixtAndBetween,
            ObjectiveType.FIGHT,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Stop the 1000 Heartless",
            hollowbastion.CheckLocation.ThousandHeartless,
            ObjectiveType.FIGHT,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Defeat Stormrider",
            landofdragons.CheckLocation.StormRiderBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Defeat Xaldin",
            beastscastle.CheckLocation.XaldinBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Defeat Future Pete",
            disneycastle.CheckLocation.FuturePete,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Defeat Grim Reaper II",
            portroyal.CheckLocation.GrimReaper2,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Defeat Genie Jafar",
            agrabah.CheckLocation.GenieJafarBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Defeat the Experiment",
            halloweentown.CheckLocation.Experiment,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Defeat Groundshaker",
            pridelands.CheckLocation.Groundshaker,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Defeat MCP",
            spaceparanoids.CheckLocation.McpBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Defeat Xemnas",
            worldthatneverwas.CheckLocation.Xemnas1Bonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Defeat Hades",
            olympuscoliseum.CheckLocation.Hades,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Help Pooh out of the pot of hunny",
            hundredacrewood.CheckLocation.StarryHillCureElement,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Participate the New Day Musical",
            atlantica.CheckLocation.MusicalBlizzardElement,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.LATEST,
        ),
    ]


def get_full_objective_list():
    from .location import (
        simulatedtwilighttown,
        twilighttown,
        hollowbastion,
        landofdragons,
        beastscastle,
        disneycastle,
        portroyal,
        agrabah,
        halloweentown,
        pridelands,
        spaceparanoids,
        worldthatneverwas,
        olympuscoliseum,
        hundredacrewood,
        atlantica,
        formlevel,
        puzzlereward,
    )

    return get_last_story_objectives() + [
        # STT
        KH2Objective(
            "Defeat Twilight Thorn",
            simulatedtwilighttown.CheckLocation.TwilightThorn,
            ObjectiveType.BOSS,
        ),
        KH2Objective(
            "Defeat Axel I",
            simulatedtwilighttown.CheckLocation.Axel1,
            ObjectiveType.FIGHT,
        ),
        KH2Objective(
            "Fight Setzer",
            simulatedtwilighttown.CheckLocation.StruggleTrophy,
            ObjectiveType.FIGHT,
        ),
        # TT
        KH2Objective(
            "Talk to the 3 Fairies",
            twilighttown.CheckLocation.ValorForm,
            ObjectiveType.WORLDPROGRESS,
        ),
        KH2Objective(
            "Defeat Sandlot Berserkers",
            twilighttown.CheckLocation.SeifersTrophy,
            ObjectiveType.FIGHT,
            ObjectiveDifficulty.MIDDLE,
        ),
        # HB
        KH2Objective(
            "Defend the Bailey", hollowbastion.CheckLocation.Bailey, ObjectiveType.FIGHT
        ),
        KH2Objective(
            "Defeat Demyx",
            hollowbastion.CheckLocation.DemyxHollowBastion,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.MIDDLE,
        ),
        # CoR
        KH2Objective(
            "Reach the end of Cavern of Remembrance",
            hollowbastion.CheckLocation.CorMineshaftUpperLevelApBoost,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Reach the end of Transport of Remembrance",
            hollowbastion.CheckLocation.TransportToRemembrance,
            ObjectiveType.FIGHT,
            ObjectiveDifficulty.LATEST,
        ),
        # LoD
        KH2Objective(
            "Climb the Mountain Trail",
            landofdragons.CheckLocation.VillageCaveAreaMap,
            ObjectiveType.WORLDPROGRESS,
        ),
        KH2Objective(
            "Fight the Enemies in Village Cave",
            landofdragons.CheckLocation.VillageCaveBonus,
            ObjectiveType.FIGHT,
        ),
        KH2Objective(
            "Defeat Shan Yu",
            landofdragons.CheckLocation.ShanYuBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.MIDDLE,
        ),
        # BC
        KH2Objective(
            "Defeat Thresholder",
            beastscastle.CheckLocation.Thresholder,
            ObjectiveType.BOSS,
        ),
        KH2Objective(
            "Help Beast", beastscastle.CheckLocation.Beast, ObjectiveType.WORLDPROGRESS
        ),
        KH2Objective(
            "Defeat Dark Thorn",
            beastscastle.CheckLocation.DarkThorn,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.MIDDLE,
        ),
        # DC
        KH2Objective(
            "Escort Queen Minnie",
            disneycastle.CheckLocation.MinnieEscort,
            ObjectiveType.WORLDPROGRESS,
        ),
        KH2Objective(
            "Fight through Windows of Time",
            disneycastle.CheckLocation.WindowOfTimeMap,
            ObjectiveType.FIGHT,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Stop Pete from Escaping",
            disneycastle.CheckLocation.BoatPete,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.MIDDLE,
        ),
        # PR
        KH2Objective(
            "Stall for time on Isla de Muerta",
            portroyal.CheckLocation.IslaDeMuertaMap,
            ObjectiveType.WORLDPROGRESS,
        ),
        KH2Objective(
            "Defend the Interceptor from pirates",
            portroyal.CheckLocation.BoatFight,
            ObjectiveType.FIGHT,
        ),
        KH2Objective(
            "Stop the explosive barrels",
            portroyal.CheckLocation.InterceptorBarrels,
            ObjectiveType.WORLDPROGRESS,
        ),
        KH2Objective(
            "Defeat Barbossa",
            portroyal.CheckLocation.Barbossa,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Defeat Grim Reaper I",
            portroyal.CheckLocation.GrimReaper1,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.MIDDLE,
        ),
        # AG
        KH2Objective(
            "Escort Abu", agrabah.CheckLocation.AbuEscort, ObjectiveType.WORLDPROGRESS
        ),
        KH2Objective(
            "Survive the Treasure Room Ambush",
            agrabah.CheckLocation.TreasureRoomBonus,
            ObjectiveType.FIGHT,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Defeat the Elemental Lords",
            agrabah.CheckLocation.ElementalLordsBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.MIDDLE,
        ),
        # HT
        KH2Objective(
            "Defeat Prison Keeper",
            halloweentown.CheckLocation.PrisonKeeper,
            ObjectiveType.BOSS,
        ),
        KH2Objective(
            "Defeat Oogie Boogie",
            halloweentown.CheckLocation.OogieBoogie,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Capture Lock Shock and Barrel",
            halloweentown.CheckLocation.LockShockBarrel,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Find the stolen presents",
            halloweentown.CheckLocation.Present,
            ObjectiveType.FIGHT,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Create decoy presents",
            halloweentown.CheckLocation.DecoyPresents,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.MIDDLE,
        ),
        # PL
        KH2Objective(
            "Reunite with Simba",
            pridelands.CheckLocation.CircleOfLife,
            ObjectiveType.WORLDPROGRESS,
        ),
        KH2Objective(
            "Rescue Timon and Pumbaa",
            pridelands.CheckLocation.Hyenas1,
            ObjectiveType.WORLDPROGRESS,
        ),
        KH2Objective(
            "Defeat Scar",
            pridelands.CheckLocation.Scar,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Get info about the ghost of Scar from Hyenas",
            pridelands.CheckLocation.Hyenas2,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.MIDDLE,
        ),
        # SP
        KH2Objective(
            "Survive the Dataspace Attack",
            spaceparanoids.CheckLocation.ScreensBonus,
            ObjectiveType.FIGHT,
        ),
        KH2Objective(
            "Defeat Hostile Program",
            spaceparanoids.CheckLocation.HostileProgramBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Ride the Solar Sailor",
            spaceparanoids.CheckLocation.SolarSailerBonus,
            ObjectiveType.FIGHT,
            ObjectiveDifficulty.MIDDLE,
        ),
        # TWTNW
        KH2Objective(
            "Defeat Roxas", worldthatneverwas.CheckLocation.Roxas, ObjectiveType.BOSS
        ),
        KH2Objective(
            "Defeat Xigbar",
            worldthatneverwas.CheckLocation.XigbarBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Defeat Luxord",
            worldthatneverwas.CheckLocation.LuxordBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Defeat Saix",
            worldthatneverwas.CheckLocation.SaixBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.MIDDLE,
        ),
        # OC
        KH2Objective(
            "Defeat Cerberus",
            olympuscoliseum.CheckLocation.Cerberus,
            ObjectiveType.BOSS,
        ),
        KH2Objective(
            "Train with Phil",
            olympuscoliseum.CheckLocation.Urns,
            ObjectiveType.WORLDPROGRESS,
        ),
        KH2Objective(
            "Defeat Pete at The Lock",
            olympuscoliseum.CheckLocation.PeteOlympusColiseum,
            ObjectiveType.FIGHT,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Defeat Hydra",
            olympuscoliseum.CheckLocation.Hydra,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Defeat the Ambush in Hades Chamber",
            olympuscoliseum.CheckLocation.AuronssStatue,
            ObjectiveType.FIGHT,
            ObjectiveDifficulty.MIDDLE,
        ),
        # Cups
        KH2Objective(
            "Win the Pain and Panic Cup",
            olympuscoliseum.CheckLocation.PainPanicCupProtectBelt,
            ObjectiveType.FIGHT,
        ),
        KH2Objective(
            "Win the Cerberus Cup",
            olympuscoliseum.CheckLocation.CerberusCupRisingDragon,
            ObjectiveType.FIGHT,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Win the Titan Cup",
            olympuscoliseum.CheckLocation.TitanCupGenjiShield,
            ObjectiveType.FIGHT,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Win the Goddess of Fate Cup",
            olympuscoliseum.CheckLocation.GoddessOfFateCupFatalCrest,
            ObjectiveType.FIGHT,
            ObjectiveDifficulty.LATE,
        ),
        # HAW
        KH2Objective(
            "Rescue Pooh from the Spooky Cave",
            hundredacrewood.CheckLocation.SweetMemories,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.LATE,
        ),
        # AT
        KH2Objective(
            "Relearn to swim",
            atlantica.CheckLocation.UnderseaKingdomMap,
            ObjectiveType.WORLDPROGRESS,
        ),
        KH2Objective(
            "Defeat Ursula",
            atlantica.CheckLocation.MysteriousAbyss,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.LATE,
        ),
        # FORM
        KH2Objective(
            "Reach Valor Level 3",
            formlevel.CheckLocation.Valor3,
            ObjectiveType.WORLDPROGRESS,
        ),
        KH2Objective(
            "Reach Valor Level 5",
            formlevel.CheckLocation.Valor5,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Reach Valor Level 7",
            formlevel.CheckLocation.Valor7,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Reach Wisdom Level 3",
            formlevel.CheckLocation.Wisdom3,
            ObjectiveType.WORLDPROGRESS,
        ),
        KH2Objective(
            "Reach Wisdom Level 5",
            formlevel.CheckLocation.Wisdom5,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Reach Wisdom Level 7",
            formlevel.CheckLocation.Wisdom7,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Reach Limit Level 3",
            formlevel.CheckLocation.Limit3,
            ObjectiveType.WORLDPROGRESS,
        ),
        KH2Objective(
            "Reach Limit Level 5",
            formlevel.CheckLocation.Limit5,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Reach Limit Level 7",
            formlevel.CheckLocation.Limit7,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Reach Master Level 3",
            formlevel.CheckLocation.Master3,
            ObjectiveType.WORLDPROGRESS,
        ),
        KH2Objective(
            "Reach Master Level 5",
            formlevel.CheckLocation.Master5,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Reach Master Level 7",
            formlevel.CheckLocation.Master7,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Reach Final Level 3",
            formlevel.CheckLocation.Final3,
            ObjectiveType.WORLDPROGRESS,
        ),
        KH2Objective(
            "Reach Final Level 5",
            formlevel.CheckLocation.Final5,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Reach Final Level 7",
            formlevel.CheckLocation.Final7,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.LATEST,
        ),
        # Puzzle
        KH2Objective(
            "Complete the Awakening Puzzle",
            puzzlereward.CheckLocation.AwakeningApBoost,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Complete the Heart Puzzle",
            puzzlereward.CheckLocation.HeartSerenityCrystal,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.MIDDLE,
        ),
        KH2Objective(
            "Complete the Duality Puzzle",
            puzzlereward.CheckLocation.DualityRareDocument,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Complete the Frontier Puzzle",
            puzzlereward.CheckLocation.FrontierManifestIllusion,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Complete the Daylight Puzzle",
            puzzlereward.CheckLocation.DaylightExecutivesRing,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.LATE,
        ),
        KH2Objective(
            "Complete the Sunset Puzzle",
            puzzlereward.CheckLocation.SunsetGrandRibbon,
            ObjectiveType.WORLDPROGRESS,
            ObjectiveDifficulty.LATE,
        ),
        # Superbosses
        KH2Objective(
            "Defeat Sephiroth",
            hollowbastion.CheckLocation.SephirothBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat Lingering Will",
            disneycastle.CheckLocation.LingeringWillBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat Data Roxas",
            simulatedtwilighttown.CheckLocation.DataRoxasMagicBoost,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat Data Demyx",
            hollowbastion.CheckLocation.DataDemyxApBoost,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat Data Xigbar",
            landofdragons.CheckLocation.DataXigbarDefenseBoost,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat Data Saix",
            pridelands.CheckLocation.DataSaix,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat Data Axel",
            twilighttown.CheckLocation.DataAxelMagicBoost,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat Data Xaldin",
            beastscastle.CheckLocation.DataXaldin,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat Data Luxord",
            portroyal.CheckLocation.DataLuxordApBoost,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat Data Xemnas",
            worldthatneverwas.CheckLocation.DataXemnas,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat Data Zexion",
            olympuscoliseum.CheckLocation.DataZexionLostIllusion,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat Data Vexen",
            halloweentown.CheckLocation.DataVexen,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat Data Larxene",
            spaceparanoids.CheckLocation.DataLarxeneLostIllusion,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat Data Lexaeus",
            agrabah.CheckLocation.DataLexaeus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat Data Marluxia",
            disneycastle.CheckLocation.DataMarluxiaLostIllusion,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        # AS
        KH2Objective(
            "Defeat AS Zexion",
            olympuscoliseum.CheckLocation.ZexionBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat AS Vexen",
            halloweentown.CheckLocation.VexenBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat AS Larxene",
            spaceparanoids.CheckLocation.LarxeneBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat AS Lexaeus",
            agrabah.CheckLocation.LexaeusBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
        KH2Objective(
            "Defeat AS Marluxia",
            disneycastle.CheckLocation.MarluxiaBonus,
            ObjectiveType.BOSS,
            ObjectiveDifficulty.LATEST,
        ),
    ]



import json
import os
import random
import string

from collections import namedtuple

from .Class.seedSettings import ExtraConfigurationData, SeedSettings
from .Module import appconfig
from .Module.RandomizerSettings import RandomizerSettings
from .Module.generate import generateSeedCLI
from .Module.seedshare import SharedSeed
from .Module.version import LOCAL_UI_VERSION

SeedInfo = namedtuple(
    "SeedInfo",
    ["seed_name", "requested_preset", "generator_string", "hash_icons", "spoiler_html"],
)


def make_random_seed_from_preset_name(requested_type: str):
    # TODO: Search for the specific preset file instead of loading all of them
    # load preset data for all presets in folder
    preset_json = {}
    for preset_file_name in os.listdir(appconfig.PRESET_FOLDER):
        preset_name, extension = os.path.splitext(preset_file_name)
        if extension == ".json":
            with open(
                os.path.join(appconfig.PRESET_FOLDER, preset_file_name), "r"
            ) as presetData:
                try:
                    settings_json = json.load(presetData)
                    preset_json[preset_name] = settings_json
                except Exception:
                    print(
                        "Unable to load preset [{}], skipping".format(preset_file_name)
                    )

    # get seed name at random
    characters = string.ascii_letters + string.digits
    seedString = "".join(random.choice(characters) for i in range(30))
    makeSpoilerLog = False
    settings = SeedSettings()
    try:
        settings.apply_settings_json(preset_json[requested_type])
    except KeyError:
        print(
            '\033[31mNo preset file found with the name \033[1m"{}"\033[0m'.format(
                requested_type
            )
        )
        exit(128)

    shared_seed = SharedSeed(
        generator_version=LOCAL_UI_VERSION,
        seed_name=seedString,
        spoiler_log=makeSpoilerLog,
        settings_string=settings.settings_string(),
        tourney_gen=True,
    )
    shared_string_text = shared_seed.to_share_string()

    rando_settings = RandomizerSettings(
        seedString, makeSpoilerLog, LOCAL_UI_VERSION, settings, shared_string_text
    )

    extra_data = ExtraConfigurationData(
        platform="PC", tourney=True, custom_cosmetics_executables=[]
    )

    spoiler_log = generateSeedCLI(rando_settings, extra_data)

    return SeedInfo(
        seed_name=seedString,
        requested_preset=requested_type,
        generator_string=shared_string_text,
        hash_icons=rando_settings.seedHashIcons,
        spoiler_html=spoiler_log,
    )



from enum import Enum

from ..configDict import locationType, itemType
from ..inventory import keyblade, ability, magic, report
from .graph import (
    DefaultLogicGraph,
    RequirementEdge,
    chest,
    popup,
    hybrid_bonus,
    stat_bonus,
    double_bonus,
    LocationGraphBuilder,
    START_NODE,
)
from ...Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    FragmentCrossing = "Fragment Crossing"
    FragmentCrossingChests = "Fragment Crossing Chests"
    Roxas = "Roxas"
    MemorysSkyscraper = "Memory's Skyscraper"
    MemorysSkyscraperChests = "Memory's Skyscraper Chests"
    BrinkOfDespair = "Brink of Despair"
    BrinkOfDespairChests = "Brink of Despair Chests"
    NothingsCall = "Nothing's Call"
    NothingsCallChests = "Nothing's Call Chests"
    TwilightsView = "Twilight's View"
    TwilightsViewChests = "Twilight's View Chests"
    Xigbar = "Xigbar"
    NaughtsSkyway = "Naught's Skyway"
    NaughtsSkywayChests = "Naught's Skyway Chests"
    Oblivion = "Oblivion"
    Luxord = "Luxord"
    Saix = "Saix"
    PreXemnas1Popup = "Pre-Xemnas 1 Popup"
    RuinAndCreationsPassage = "Ruin and Creation's Passage"
    RuinAndCreationsPassageChests = "Ruin and Creation's Passage Chests"
    Xemnas1 = "Xemnas 1"
    FinalXemnas = "Final Xemnas"
    DataXemnas = "Data Xemnas"


class CheckLocation(str, Enum):
    FragmentCrossingMythrilStone = "Fragment Crossing Mythril Stone"
    FragmentCrossingMythrilCrystal = "Fragment Crossing Mythril Crystal"
    FragmentCrossingApBoost = "Fragment Crossing AP Boost"
    FragmentCrossingOrichalcum = "Fragment Crossing Orichalcum"
    Roxas = "Roxas"
    RoxasSecretAnsemReport8 = "Roxas Secret Ansem Report 8"
    TwoBecomeOne = "Two Become One"
    MemorysSkyscraperMythrilCrystal = "Memory's Skyscaper Mythril Crystal"
    MemorysSkyscraperApBoost = "Memory's Skyscaper AP Boost"
    MemorysSkyscraperMythrilStone = "Memory's Skyscaper Mythril Stone"
    BrinkOfDespairDarkCityMap = "The Brink of Despair Dark City Map"
    BrinkOfDespairOrichalcumPlus = "The Brink of Despair Orichalcum+"
    NothingsCallMythrilGem = "Nothing's Call Mythril Gem"
    NothingsCallOrichalcum = "Nothing's Call Orichalcum"
    TwilightsViewCosmicBelt = "Twilight's View Cosmic Belt"
    XigbarBonus = "Xigbar Bonus"
    XigbarSecretAnsemReport3 = "Xigbar Secret Ansem Report 3"
    NaughtsSkywayMythrilGem = "Naught's Skyway Mythril Gem"
    NaughtsSkywayOrichalcum = "Naught's Skyway Orichalcum"
    NaughtsSkywayMythrilCrystal = "Naught's Skyway Mythril Crystal"
    Oblivion = "Oblivion"
    CastleThatNeverWasMap = "Castle That Never Was Map"
    LuxordBonus = "Luxord Bonus"
    LuxordSecretAnsemReport9 = "Luxord Secret Ansem Report 9"
    SaixBonus = "Saix Bonus"
    SaixSecretAnsemReport12 = "Saix Secret Ansem Report 12"
    PreXemnas1SecretAnsemReport11 = "(Pre-Xemnas 1) Secret Ansem Report 11"
    RuinCreationsPassageMythrilStone = "Ruin and Creation's Passage Mythril Stone"
    RuinCreationsPassageApBoost = "Ruin and Creation's Passage AP Boost"
    RuinCreationsPassageMythrilCrystal = "Ruin and Creation's Passage Mythril Crystal"
    RuinCreationsPassageOrichalcum = "Ruin and Creation's Passage Orichalcum"
    Xemnas1Bonus = "Xemnas 1 Bonus"
    Xemnas1SecretAnsemReport13 = "Xemnas 1 Secret Ansem Report 13"
    FinalXemnas = "Final Xemnas"
    DataXemnas = "Xemnas (Data) Power Boost"


class TWTNWLogicGraph(DefaultLogicGraph):
    def __init__(self, reverse_rando, keyblade_unlocks, objectives_needed):
        DefaultLogicGraph.__init__(self, NodeId)
        keyblade_lambda = (
            lambda inv: not keyblade_unlocks
            or ItemPlacementHelpers.need_twtnw_keyblade(inv)
        )
        self.logic[NodeId.FragmentCrossing][
            NodeId.FragmentCrossingChests
        ] = keyblade_lambda
        self.logic[NodeId.MemorysSkyscraper][
            NodeId.MemorysSkyscraperChests
        ] = keyblade_lambda
        self.logic[NodeId.BrinkOfDespair][NodeId.BrinkOfDespairChests] = keyblade_lambda
        self.logic[NodeId.NothingsCall][NodeId.NothingsCallChests] = keyblade_lambda
        self.logic[NodeId.TwilightsView][NodeId.TwilightsViewChests] = keyblade_lambda
        self.logic[NodeId.NaughtsSkyway][NodeId.NaughtsSkywayChests] = keyblade_lambda
        self.logic[NodeId.RuinAndCreationsPassage][
            NodeId.RuinAndCreationsPassageChests
        ] = keyblade_lambda

        completion_lambda = lambda inv: ItemPlacementHelpers.need_proofs(inv)
        if objectives_needed > 0:
            completion_lambda = ItemPlacementHelpers.make_need_objectives_lambda(
                objectives_needed
            )

        self.logic[START_NODE][NodeId.FinalXemnas] = (
            lambda inv: ItemPlacementHelpers.need_promise_charm(inv)
            and completion_lambda(inv)
        )
        if not reverse_rando:
            self.logic[NodeId.FragmentCrossing][
                NodeId.Roxas
            ] = ItemPlacementHelpers.twtnw_roxas_check
            self.logic[NodeId.Saix][
                NodeId.PreXemnas1Popup
            ] = ItemPlacementHelpers.twtnw_post_saix_check
            self.logic[NodeId.Xemnas1][NodeId.FinalXemnas] = completion_lambda
        else:
            self.logic[NodeId.FragmentCrossing][
                NodeId.Xemnas1
            ] = ItemPlacementHelpers.twtnw_roxas_check
            self.logic[NodeId.Xigbar][
                NodeId.PreXemnas1Popup
            ] = ItemPlacementHelpers.twtnw_post_saix_check
            self.logic[NodeId.Roxas][NodeId.FinalXemnas] = completion_lambda


def make_graph(graph: LocationGraphBuilder):
    twtnw = locationType.TWTNW
    objectives_needed = 0
    if graph.settings.objective_rando:
        objectives_needed = graph.settings.num_objectives_needed
    if graph.settings.emblems:
        objectives_needed = graph.settings.num_emblems_needed
    twtnw_logic = TWTNWLogicGraph(
        graph.reverse_rando, graph.keyblades_unlock_chests, objectives_needed
    )
    graph.add_logic(twtnw_logic)

    fragment_crossing_chests = graph.add_location(
        NodeId.FragmentCrossingChests,
        [
            chest(374, CheckLocation.FragmentCrossingMythrilStone, twtnw),
            chest(375, CheckLocation.FragmentCrossingMythrilCrystal, twtnw),
            chest(376, CheckLocation.FragmentCrossingApBoost, twtnw),
            chest(377, CheckLocation.FragmentCrossingOrichalcum, twtnw),
        ],
    )
    fragment_crossing = graph.add_location(NodeId.FragmentCrossing, [])
    roxas = graph.add_location(
        NodeId.Roxas,
        [
            hybrid_bonus(69, CheckLocation.Roxas, twtnw, vanilla=ability.ComboMaster),
            popup(
                532,
                CheckLocation.RoxasSecretAnsemReport8,
                twtnw,
                vanilla=report.AnsemReport8,
            ),
            popup(
                277, CheckLocation.TwoBecomeOne, twtnw, vanilla=keyblade.TwoBecomeOne
            ),
        ],
    )
    memorys_skyscraper_chests = graph.add_location(
        NodeId.MemorysSkyscraperChests,
        [
            chest(391, CheckLocation.MemorysSkyscraperMythrilCrystal, twtnw),
            chest(523, CheckLocation.MemorysSkyscraperApBoost, twtnw),
            chest(524, CheckLocation.MemorysSkyscraperMythrilStone, twtnw),
        ],
    )
    memorys_skyscraper = graph.add_location(NodeId.MemorysSkyscraper, [])
    brink_of_despair_chests = graph.add_location(
        NodeId.BrinkOfDespairChests,
        [
            chest(335, CheckLocation.BrinkOfDespairDarkCityMap, twtnw),
            chest(500, CheckLocation.BrinkOfDespairOrichalcumPlus, twtnw),
        ],
    )
    brink_of_despair = graph.add_location(NodeId.BrinkOfDespair, [])
    nothings_call_chests = graph.add_location(
        NodeId.NothingsCallChests,
        [
            chest(378, CheckLocation.NothingsCallMythrilGem, twtnw),
            chest(379, CheckLocation.NothingsCallOrichalcum, twtnw),
        ],
    )
    nothings_call = graph.add_location(NodeId.NothingsCall, [])
    twilights_view_chests = graph.add_location(
        NodeId.TwilightsViewChests,
        [
            chest(336, CheckLocation.TwilightsViewCosmicBelt, twtnw),
        ],
    )
    twilights_view = graph.add_location(NodeId.TwilightsView, [])
    xigbar = graph.add_location(
        NodeId.Xigbar,
        [
            stat_bonus(23, CheckLocation.XigbarBonus, twtnw),
            popup(
                527,
                CheckLocation.XigbarSecretAnsemReport3,
                twtnw,
                vanilla=report.AnsemReport3,
            ),
        ],
    )
    naughts_skyway_chests = graph.add_location(
        NodeId.NaughtsSkywayChests,
        [
            chest(380, CheckLocation.NaughtsSkywayMythrilGem, twtnw),
            chest(381, CheckLocation.NaughtsSkywayOrichalcum, twtnw),
            chest(382, CheckLocation.NaughtsSkywayMythrilCrystal, twtnw),
        ],
    )
    naughts_skyway = graph.add_location(NodeId.NaughtsSkyway, [])
    oblivion = graph.add_location(
        NodeId.Oblivion,
        [
            popup(278, CheckLocation.Oblivion, twtnw, vanilla=keyblade.Oblivion),
            popup(496, CheckLocation.CastleThatNeverWasMap, twtnw),
        ],
    )
    luxord = graph.add_location(
        NodeId.Luxord,
        [
            hybrid_bonus(24, CheckLocation.LuxordBonus, twtnw, vanilla=magic.Magnet),
            popup(
                533,
                CheckLocation.LuxordSecretAnsemReport9,
                twtnw,
                vanilla=report.AnsemReport9,
            ),
        ],
    )
    saix = graph.add_location(
        NodeId.Saix,
        [
            stat_bonus(25, CheckLocation.SaixBonus, twtnw),
            popup(
                536,
                CheckLocation.SaixSecretAnsemReport12,
                twtnw,
                vanilla=report.AnsemReport12,
            ),
        ],
    )
    pre_xemnas_1_popup = graph.add_location(
        NodeId.PreXemnas1Popup,
        [
            popup(
                535,
                CheckLocation.PreXemnas1SecretAnsemReport11,
                twtnw,
                vanilla=report.AnsemReport11,
            ),
        ],
    )
    ruin_creations_passage_chests = graph.add_location(
        NodeId.RuinAndCreationsPassageChests,
        [
            chest(385, CheckLocation.RuinCreationsPassageMythrilStone, twtnw),
            chest(386, CheckLocation.RuinCreationsPassageApBoost, twtnw),
            chest(387, CheckLocation.RuinCreationsPassageMythrilCrystal, twtnw),
            chest(388, CheckLocation.RuinCreationsPassageOrichalcum, twtnw),
        ],
    )
    ruin_creations_passage = graph.add_location(NodeId.RuinAndCreationsPassage, [])
    xemnas_1 = graph.add_location(
        NodeId.Xemnas1,
        [
            double_bonus(26, CheckLocation.Xemnas1Bonus, twtnw),
            popup(
                537,
                CheckLocation.Xemnas1SecretAnsemReport13,
                twtnw,
                vanilla=report.AnsemReport13,
            ),
        ],
    )
    final_xemnas = graph.add_location(
        NodeId.FinalXemnas,
        [
            stat_bonus(
                71,
                CheckLocation.FinalXemnas,
                twtnw,
                invalid_checks=[
                    e
                    for e in itemType
                    if e
                    not in [
                        itemType.GAUGE,
                        itemType.SLOT,
                        itemType.SYNTH,
                        itemType.ITEM,
                    ]
                ],
            ),
        ],
    )
    data_xemnas = graph.add_location(
        NodeId.DataXemnas,
        [
            popup(554, CheckLocation.DataXemnas, [twtnw, locationType.DataOrg]),
        ],
    )

    graph.register_superboss(data_xemnas)

    graph.add_edge(START_NODE, final_xemnas, RequirementEdge(battle=True, strict=False))

    graph.add_edge(fragment_crossing, fragment_crossing_chests)
    graph.add_edge(memorys_skyscraper, memorys_skyscraper_chests)
    graph.add_edge(brink_of_despair, brink_of_despair_chests)
    graph.add_edge(nothings_call, nothings_call_chests)
    graph.add_edge(twilights_view, twilights_view_chests)
    graph.add_edge(naughts_skyway, naughts_skyway_chests)
    graph.add_edge(ruin_creations_passage, ruin_creations_passage_chests)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, fragment_crossing)
        graph.add_edge(fragment_crossing, roxas, RequirementEdge(battle=True))
        graph.add_edge(roxas, memorys_skyscraper)
        graph.add_edge(memorys_skyscraper, brink_of_despair)
        graph.add_edge(brink_of_despair, nothings_call)
        graph.add_edge(nothings_call, twilights_view)
        graph.add_edge(twilights_view, xigbar, RequirementEdge(battle=True))
        graph.add_edge(xigbar, naughts_skyway)
        graph.add_edge(naughts_skyway, oblivion)
        graph.add_edge(oblivion, luxord, RequirementEdge(battle=True))
        graph.add_edge(luxord, saix, RequirementEdge(battle=True))
        graph.add_edge(saix, pre_xemnas_1_popup)
        graph.add_edge(pre_xemnas_1_popup, ruin_creations_passage)
        graph.add_edge(ruin_creations_passage, xemnas_1, RequirementEdge(battle=True))
        graph.add_edge(
            xemnas_1, final_xemnas, RequirementEdge(battle=True, strict=False)
        )
        graph.add_edge(xemnas_1, data_xemnas, RequirementEdge(battle=True))
        graph.register_first_boss(saix)
        graph.register_last_story_boss(xemnas_1)
    else:
        graph.add_edge(START_NODE, fragment_crossing)
        graph.add_edge(fragment_crossing, xemnas_1, RequirementEdge(battle=True))
        graph.add_edge(xemnas_1, memorys_skyscraper)
        graph.add_edge(memorys_skyscraper, brink_of_despair)
        graph.add_edge(brink_of_despair, nothings_call)
        graph.add_edge(nothings_call, twilights_view)
        graph.add_edge(twilights_view, saix, RequirementEdge(battle=True))
        graph.add_edge(saix, naughts_skyway)
        graph.add_edge(naughts_skyway, oblivion)
        graph.add_edge(oblivion, luxord, RequirementEdge(battle=True))
        graph.add_edge(luxord, xigbar, RequirementEdge(battle=True))
        graph.add_edge(xigbar, pre_xemnas_1_popup)
        graph.add_edge(pre_xemnas_1_popup, ruin_creations_passage)
        graph.add_edge(ruin_creations_passage, roxas, RequirementEdge(battle=True))
        graph.add_edge(roxas, final_xemnas, RequirementEdge(battle=True, strict=False))
        graph.add_edge(roxas, data_xemnas, RequirementEdge(battle=True))
        graph.register_first_boss(xigbar)
        graph.register_last_story_boss(roxas)



from enum import Enum

from ..configDict import locationType, itemType
from ..inventory import (
    magic,
    keyblade,
    ability,
    summon,
    report,
    storyunlock,
    form,
    misc,
)
from .graph import (
    DefaultLogicGraph,
    RequirementEdge,
    chest,
    popup,
    hybrid_bonus,
    stat_bonus,
    item_bonus,
    LocationGraphBuilder,
    START_NODE,
)
from ...Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    MarketplaceMapPopup = "Marketplace Map Popup"
    Borough = "Borough"
    BoroughChests = "Borough Chests"
    MerlinsHousePopup = "Merlin's House Popup"
    Bailey = "Bailey"
    BaseballCharmPopup = "Baseball Charm Popup"
    Postern = "Postern"
    PosternChests = "Postern Chests"
    Corridors = "Cooridors"
    CorridorsChests = "Cooridors Chests"
    DoorToDarknessPopup = "DTD Popup"
    AnsemsStudy = "Ansem's Study"
    AnsemsStudyChests = "Ansem's Study Chests"
    UkuleleCharm = "Ukulele Charm"
    UkuleleCharmChests = "Ukulele Charm Chests"
    RestorationSite = "Restoration Site"
    RestorationSiteChests = "Restoration Site Chests"
    DemyxHollowBastion = "Demyx (HB)"
    FinalFantasyFights = "Final Fantasy Fights"
    CrystalFissure = "Crystal Fissure"
    CrystalFissureChests = "Crystal Fissure Chests"
    ThousandHeartless = "1000 Heartless"
    GullWing = "Gull Wing"
    GullWingChests = "Gull Wing Chests"
    HeartlessManufactory = "Heartless Manufactory"
    HeartlessManufactoryChests = "Heartless Manufactory Chests"
    Sephiroth = "Sephiroth"
    Mushroom13 = "Mushroom 13"
    DataDemyx = "Data Demyx"
    CorDepths = "CoR Depths"
    CorDepthsChests = "CoR Depths Chests"
    CorMineshaftPreFigh1 = "CoR Mineshaft Pre-Fight 1"
    CorMineshaftPreFigh1Chests = "CoR Mineshaft Pre-Fight 1 Chests"
    CorDepthsPostFight1 = "CoR-Depths Post-Fight 1"
    CorDepthsPostFight1Chests = "CoR-Depths Post-Fight 1 Chests"
    CorMiningArea = "CoR Mining Area"
    CorMiningAreaChests = "CoR Mining Area Chests"
    CorMineshaftPreFight2 = "CoR Minshaft Pre-Fight 2"
    CorMineshaftPreFight2Chests = "CoR Minshaft Pre-Fight 2 Chests"
    CorEngineChamber = "CoR Engine Chamber"
    CorEngineChamberChests = "CoR Engine Chamber Chests"
    CorMineshaftPostFight2 = "CoR Mineshaft Post Fight 2"
    CorMineshaftPostFight2Chests = "CoR Mineshaft Post Fight 2 Chests"
    CorMineshaftLastChest = "CoR Mineshaft Last Chest"
    CorMineshaftLastChestChests = "CoR Mineshaft Last Chest Chests"
    TransportToRemembrance = "Transport to Remembrance"


class CheckLocation(str, Enum):
    MarketplaceMap = "Marketplace Map"
    BoroughDriveRecovery = "Borough Drive Recovery"
    BoroughApBoost = "Borough AP Boost"
    BoroughHiPotion = "Borough Hi-Potion"
    BoroughMythrilShard = "Borough Mythril Shard"
    BoroughDarkShard = "Borough Dark Shard"
    MerlinsHouseMembershipCard = "Merlin's House Membership Card"
    MerlinsHouseBlizzardElement = "Merlin's House Blizzard Element"
    Bailey = "Bailey"
    BaileySecretAnsemReport7 = "Bailey Secret Ansem Report 7"
    BaseballCharm = "Baseball Charm"
    PosternCastlePerimeterMap = "Postern Castle Perimeter Map"
    PosternMythrilGem = "Postern Mythril Gem"
    PosternApBoost = "Postern AP Boost"
    CorridorsMythrilStone = "Corridors Mythril Stone"
    CorridorsMythrilCrystal = "Corridors Mythril Crystal"
    CorridorsDarkCrystal = "Corridors Dark Crystal"
    CorridorsApBoost = "Corridors AP Boost"
    AnsemsStudyMasterForm = "Ansem's Study Master Form"
    AnsemsStudySleepingLion = "Ansem's Study Sleeping Lion"
    AnsemsStudySkillRecipe = "Ansem's Study Skill Recipe"
    UkuleleCharm = "Ansem's Study Ukulele Charm"
    RestorationSiteMoonRecipe = "Restoration Site Moon Recipe"
    RestorationSiteApBoost = "Restoration Site AP Boost"
    DemyxHollowBastion = "Demyx (HB)"
    FinalFantasyFightsCureElement = "FF Fights Cure Element"
    CrystalFissureTornPages = "Crystal Fissure Torn Pages"
    CrystalFissureGreatMawMap = "Crystal Fissure The Great Maw Map"
    CrystalFissureEnergyCrystal = "Crystal Fissure Energy Crystal"
    CrystalFissureApBoost = "Crystal Fissure AP Boost"
    ThousandHeartless = "1000 Heartless"
    ThousandHeartlessSecretAnsemReport1 = "1000 Heartless Secret Ansem Report 1"
    ThousandHeartlessIceCream = "1000 Heartless Ice Cream"
    ThousandHeartlessPicture = "1000 Heartless Picture"
    PosternGullWing = "Postern Gull Wing"
    HeartlessManufactoryCosmicChain = "Heartless Manufactory Cosmic Chain"
    SephirothBonus = "Sephiroth Bonus"
    SephirothFenrir = "Sephiroth Fenrir"
    WinnersProof = "Winner's Proof"
    ProofOfPeace = "Proof of Peace"
    DataDemyxApBoost = "Demyx (Data) AP Boost"

    CorDepthsApBoost = "CoR Depths AP Boost"
    CorDepthsPowerCrystal = "CoR Depths Power Crystal"
    CorDepthsFrostCrystal = "CoR Depths Frost Crystal"
    CorDepthsManifestIllusion = "CoR Depths Manifest Illusion"
    CorDepthsApBoost2 = "CoR Depths AP Boost (2)"
    CorMineshaftLowerLevelDepthsOfRemembranceMap = (
        "CoR Mineshaft Lower Level Depths of Remembrance Map"
    )
    CorMineshaftLowerLevelApBoost = "CoR Mineshaft Lower Level AP Boost"
    CorDepthsUpperLevelRemembranceGem = "CoR Depths Upper Level Remembrance Gem"
    CorMiningAreaSerenityGem = "CoR Mining Area Serenity Gem"
    CorMiningAreaApBoost = "CoR Mining Area AP Boost"
    CorMiningAreaSerenityCrystal = "CoR Mining Area Serenity Crystal"
    CorMiningAreaManifestIllusion = "CoR Mining Area Manifest Illusion"
    CorMiningAreaSerenityGem2 = "CoR Mining Area Serenity Gem (2)"
    CorMiningAreaDarkRemembranceMap = "CoR Mining Area Dark Remembrance Map"
    CorMineshaftMidLevelPowerBoost = "CoR Mineshaft Mid Level Power Boost"
    CorEngineChamberSerenityCrystal = "CoR Engine Chamber Serenity Crystal"
    CorEngineChamberRemembranceCrystal = "CoR Engine Chamber Remembrance Crystal"
    CorEngineChamberApBoost = "CoR Engine Chamber AP Boost"
    CorEngineChamberManifestIllusion = "CoR Engine Chamber Manifest Illusion"
    CorMineshaftUpperLevelMagicBoost = "CoR Mineshaft Upper Level Magic Boost"
    CorMineshaftUpperLevelApBoost = "CoR Mineshaft Upper Level AP Boost"
    TransportToRemembrance = "Transport to Remembrance"


class HBLogicGraph(DefaultLogicGraph):
    def __init__(self, reverse_rando, keyblade_unlocks):
        DefaultLogicGraph.__init__(self, NodeId)
        hb_keyblade_lambda = (
            lambda inv: not keyblade_unlocks
            or ItemPlacementHelpers.need_hb_keyblade(inv)
        )
        cor_keyblade_lambda = (
            lambda inv: not keyblade_unlocks
            or ItemPlacementHelpers.need_cor_keyblade(inv)
        )
        self.logic[NodeId.Borough][NodeId.BoroughChests] = hb_keyblade_lambda
        self.logic[NodeId.Postern][NodeId.PosternChests] = hb_keyblade_lambda
        self.logic[NodeId.Corridors][NodeId.CorridorsChests] = hb_keyblade_lambda
        self.logic[NodeId.AnsemsStudy][NodeId.AnsemsStudyChests] = hb_keyblade_lambda
        self.logic[NodeId.UkuleleCharm][NodeId.UkuleleCharmChests] = hb_keyblade_lambda
        self.logic[NodeId.RestorationSite][
            NodeId.RestorationSiteChests
        ] = hb_keyblade_lambda
        self.logic[NodeId.CrystalFissure][
            NodeId.CrystalFissureChests
        ] = hb_keyblade_lambda
        self.logic[NodeId.GullWing][NodeId.GullWingChests] = hb_keyblade_lambda
        self.logic[NodeId.HeartlessManufactory][
            NodeId.HeartlessManufactoryChests
        ] = hb_keyblade_lambda

        self.logic[NodeId.CorDepths][NodeId.CorDepthsChests] = cor_keyblade_lambda
        self.logic[NodeId.CorMineshaftPreFigh1][
            NodeId.CorMineshaftPreFigh1Chests
        ] = cor_keyblade_lambda
        self.logic[NodeId.CorDepthsPostFight1][
            NodeId.CorDepthsPostFight1Chests
        ] = cor_keyblade_lambda
        self.logic[NodeId.CorMiningArea][
            NodeId.CorMiningAreaChests
        ] = cor_keyblade_lambda
        self.logic[NodeId.CorMineshaftPreFight2][
            NodeId.CorMineshaftPreFight2Chests
        ] = cor_keyblade_lambda
        self.logic[NodeId.CorEngineChamber][
            NodeId.CorEngineChamberChests
        ] = cor_keyblade_lambda
        self.logic[NodeId.CorMineshaftPostFight2][
            NodeId.CorMineshaftPostFight2Chests
        ] = cor_keyblade_lambda
        self.logic[NodeId.CorMineshaftLastChest][
            NodeId.CorMineshaftLastChestChests
        ] = cor_keyblade_lambda
        if not reverse_rando:
            self.logic[START_NODE][
                NodeId.MarketplaceMapPopup
            ] = ItemPlacementHelpers.hb1_check
            self.logic[NodeId.BaseballCharmPopup][
                NodeId.Postern
            ] = ItemPlacementHelpers.hb2_check
            self.logic[NodeId.DoorToDarknessPopup][
                NodeId.CorDepths
            ] = ItemPlacementHelpers.need_growths
            self.logic[NodeId.ThousandHeartless][
                NodeId.Mushroom13
            ] = ItemPlacementHelpers.need_proof_peace
            self.logic[NodeId.ThousandHeartless][
                NodeId.DataDemyx
            ] = ItemPlacementHelpers.need_forms
        else:
            self.logic[START_NODE][NodeId.Borough] = ItemPlacementHelpers.hb1_check
            self.logic[NodeId.ThousandHeartless][NodeId.CorDepths] = (
                lambda inv: ItemPlacementHelpers.hb2_check(inv)
                and ItemPlacementHelpers.need_growths(inv)
            )
            self.logic[NodeId.BaseballCharmPopup][
                NodeId.DataDemyx
            ] = ItemPlacementHelpers.need_forms


def make_graph(graph: LocationGraphBuilder):
    hb = locationType.HB
    cor = locationType.CoR
    hb_logic = HBLogicGraph(graph.reverse_rando, graph.keyblades_unlock_chests)
    graph.add_logic(hb_logic)

    marketplace_map_popup = graph.add_location(
        NodeId.MarketplaceMapPopup,
        [
            popup(362, CheckLocation.MarketplaceMap, hb),
        ],
    )
    borough_chests = graph.add_location(
        NodeId.BoroughChests,
        [
            chest(194, CheckLocation.BoroughDriveRecovery, hb),
            chest(195, CheckLocation.BoroughApBoost, hb),
            chest(196, CheckLocation.BoroughHiPotion, hb),
            chest(305, CheckLocation.BoroughMythrilShard, hb),
            chest(506, CheckLocation.BoroughDarkShard, hb),
        ],
    )
    borough = graph.add_location(NodeId.Borough, [])
    merlins_house_popup = graph.add_location(
        NodeId.MerlinsHousePopup,
        [
            popup(
                256,
                CheckLocation.MerlinsHouseMembershipCard,
                hb,
                vanilla=storyunlock.MembershipCard,
            ),
            popup(
                292,
                CheckLocation.MerlinsHouseBlizzardElement,
                hb,
                vanilla=magic.Blizzard,
            ),
        ],
    )
    bailey = graph.add_location(
        NodeId.Bailey,
        [
            item_bonus(47, CheckLocation.Bailey, hb, vanilla=magic.Fire),
            popup(
                531,
                CheckLocation.BaileySecretAnsemReport7,
                hb,
                vanilla=report.AnsemReport7,
            ),
        ],
    )
    baseball_charm_popup = graph.add_location(
        NodeId.BaseballCharmPopup,
        [
            popup(258, CheckLocation.BaseballCharm, hb, vanilla=summon.BaseballCharm),
        ],
    )
    postern_chests = graph.add_location(
        NodeId.PosternChests,
        [
            chest(310, CheckLocation.PosternCastlePerimeterMap, hb),
            chest(189, CheckLocation.PosternMythrilGem, hb),
            chest(190, CheckLocation.PosternApBoost, hb),
        ],
    )
    postern = graph.add_location(NodeId.Postern, [])
    corridors_chests = graph.add_location(
        NodeId.CorridorsChests,
        [
            chest(200, CheckLocation.CorridorsMythrilStone, hb),
            chest(201, CheckLocation.CorridorsMythrilCrystal, hb),
            chest(202, CheckLocation.CorridorsDarkCrystal, hb),
            chest(307, CheckLocation.CorridorsApBoost, hb),
        ],
    )
    corridors = graph.add_location(NodeId.Corridors, [])
    dtd_popup = graph.add_location(
        NodeId.DoorToDarknessPopup,
        [
            popup(
                266, CheckLocation.AnsemsStudyMasterForm, hb, vanilla=form.MasterForm
            ),
            popup(
                276,
                CheckLocation.AnsemsStudySleepingLion,
                hb,
                vanilla=keyblade.SleepingLion,
            ),
        ],
    )
    ansems_study_chests = graph.add_location(
        NodeId.AnsemsStudyChests,
        [
            chest(184, CheckLocation.AnsemsStudySkillRecipe, hb),
        ],
    )
    ansems_study = graph.add_location(NodeId.AnsemsStudy, [])
    ukulele_charm_chests = graph.add_location(
        NodeId.UkuleleCharmChests,
        [
            chest(183, CheckLocation.UkuleleCharm, hb, vanilla=summon.UkuleleCharm),
        ],
    )
    ukulele_charm = graph.add_location(NodeId.UkuleleCharm, [])
    restoration_site_chests = graph.add_location(
        NodeId.RestorationSiteChests,
        [
            chest(309, CheckLocation.RestorationSiteMoonRecipe, hb),
            chest(507, CheckLocation.RestorationSiteApBoost, hb),
        ],
    )
    restoration_site = graph.add_location(NodeId.RestorationSite, [])
    demyx = graph.add_location(
        NodeId.DemyxHollowBastion,
        [
            hybrid_bonus(
                28, CheckLocation.DemyxHollowBastion, hb, vanilla=magic.Blizzard
            ),
        ],
    )
    final_fantasy_fights = graph.add_location(
        NodeId.FinalFantasyFights,
        [
            popup(
                361, CheckLocation.FinalFantasyFightsCureElement, hb, vanilla=magic.Cure
            ),
        ],
    )
    crystal_fissure_chests = graph.add_location(
        NodeId.CrystalFissureChests,
        [
            chest(
                179, CheckLocation.CrystalFissureTornPages, hb, vanilla=misc.TornPages
            ),
            chest(489, CheckLocation.CrystalFissureGreatMawMap, hb),
            chest(180, CheckLocation.CrystalFissureEnergyCrystal, hb),
            chest(181, CheckLocation.CrystalFissureApBoost, hb),
        ],
    )
    crystal_fissure = graph.add_location(NodeId.CrystalFissure, [])
    thousand_heartless = graph.add_location(
        NodeId.ThousandHeartless,
        [
            item_bonus(
                60, CheckLocation.ThousandHeartless, hb, vanilla=ability.GuardBreak
            ),
            popup(
                525,
                CheckLocation.ThousandHeartlessSecretAnsemReport1,
                hb,
                vanilla=report.AnsemReport1,
            ),
            popup(
                269,
                CheckLocation.ThousandHeartlessIceCream,
                hb,
                vanilla=storyunlock.IceCream,
            ),
            popup(511, CheckLocation.ThousandHeartlessPicture, hb),
        ],
    )
    gull_wing_chests = graph.add_location(
        NodeId.GullWingChests,
        [
            chest(491, CheckLocation.PosternGullWing, hb, vanilla=keyblade.GullWing),
        ],
    )
    gull_wing = graph.add_location(NodeId.GullWing, [])
    heartless_manufactory_chests = graph.add_location(
        NodeId.HeartlessManufactoryChests,
        [
            chest(
                311,
                CheckLocation.HeartlessManufactoryCosmicChain,
                hb,
                invalid_checks=[itemType.MANUFACTORYUNLOCK],
            ),
        ],
    )
    heartless_manufactory = graph.add_location(NodeId.HeartlessManufactory, [])
    sephiroth = graph.add_location(
        NodeId.Sephiroth,
        [
            stat_bonus(35, CheckLocation.SephirothBonus, [hb, locationType.Sephi]),
            popup(
                282,
                CheckLocation.SephirothFenrir,
                [hb, locationType.Sephi],
                vanilla=keyblade.Fenrir,
            ),
        ],
    )
    mushroom_13 = graph.add_location(
        NodeId.Mushroom13,
        [
            popup(
                588,
                CheckLocation.WinnersProof,
                [hb, locationType.Mush13],
                invalid_checks=[itemType.PROOF_OF_PEACE],
                vanilla=keyblade.WinnersProof,
            ),
            popup(
                589,
                CheckLocation.ProofOfPeace,
                [hb, locationType.Mush13],
                invalid_checks=[itemType.PROOF_OF_PEACE],
            ),
        ],
    )
    data_demyx = graph.add_location(
        NodeId.DataDemyx,
        [
            popup(
                560,
                CheckLocation.DataDemyxApBoost,
                [hb, locationType.DataOrg],
                invalid_checks=[itemType.FORM],
            ),
        ],
    )

    cor_depths_chests = graph.add_location(
        NodeId.CorDepthsChests,
        [
            chest(562, CheckLocation.CorDepthsApBoost, [hb, cor]),
            chest(563, CheckLocation.CorDepthsPowerCrystal, [hb, cor]),
            chest(564, CheckLocation.CorDepthsFrostCrystal, [hb, cor]),
            chest(565, CheckLocation.CorDepthsManifestIllusion, [hb, cor]),
            chest(566, CheckLocation.CorDepthsApBoost2, [hb, cor]),
        ],
    )
    cor_depths = graph.add_location(NodeId.CorDepths, [])
    cor_mineshaft_pre_fight_1_chests = graph.add_location(
        NodeId.CorMineshaftPreFigh1Chests,
        [
            chest(
                580,
                CheckLocation.CorMineshaftLowerLevelDepthsOfRemembranceMap,
                [hb, cor],
            ),
            chest(578, CheckLocation.CorMineshaftLowerLevelApBoost, [hb, cor]),
        ],
    )
    cor_mineshaft_pre_fight_1 = graph.add_location(NodeId.CorMineshaftPreFigh1, [])
    cor_depths_post_fight_1_chests = graph.add_location(
        NodeId.CorDepthsPostFight1Chests,
        [
            chest(567, CheckLocation.CorDepthsUpperLevelRemembranceGem, [hb, cor]),
        ],
    )
    cor_depths_post_fight_1 = graph.add_location(NodeId.CorDepthsPostFight1, [])
    cor_mining_area_chests = graph.add_location(
        NodeId.CorMiningAreaChests,
        [
            chest(568, CheckLocation.CorMiningAreaSerenityGem, [hb, cor]),
            chest(569, CheckLocation.CorMiningAreaApBoost, [hb, cor]),
            chest(570, CheckLocation.CorMiningAreaSerenityCrystal, [hb, cor]),
            chest(571, CheckLocation.CorMiningAreaManifestIllusion, [hb, cor]),
            chest(572, CheckLocation.CorMiningAreaSerenityGem2, [hb, cor]),
            chest(573, CheckLocation.CorMiningAreaDarkRemembranceMap, [hb, cor]),
        ],
    )
    cor_mining_area = graph.add_location(NodeId.CorMiningArea, [])
    cor_mineshaft_pre_fight_2_chests = graph.add_location(
        NodeId.CorMineshaftPreFight2Chests,
        [
            chest(581, CheckLocation.CorMineshaftMidLevelPowerBoost, [hb, cor]),
        ],
    )
    cor_mineshaft_pre_fight_2 = graph.add_location(NodeId.CorMineshaftPreFight2, [])
    cor_engine_chamber_chests = graph.add_location(
        NodeId.CorEngineChamberChests,
        [
            chest(574, CheckLocation.CorEngineChamberSerenityCrystal, [hb, cor]),
            chest(575, CheckLocation.CorEngineChamberRemembranceCrystal, [hb, cor]),
            chest(576, CheckLocation.CorEngineChamberApBoost, [hb, cor]),
            chest(577, CheckLocation.CorEngineChamberManifestIllusion, [hb, cor]),
        ],
    )
    cor_engine_chamber = graph.add_location(NodeId.CorEngineChamber, [])
    cor_mineshaft_post_fight_2_chests = graph.add_location(
        NodeId.CorMineshaftPostFight2Chests,
        [
            chest(582, CheckLocation.CorMineshaftUpperLevelMagicBoost, [hb, cor]),
        ],
    )
    cor_mineshaft_post_fight_2 = graph.add_location(NodeId.CorMineshaftPostFight2, [])
    cor_mineshaft_last_chest_chests = graph.add_location(
        NodeId.CorMineshaftLastChestChests,
        [
            chest(579, CheckLocation.CorMineshaftUpperLevelApBoost, [hb, cor]),
        ],
    )
    cor_mineshaft_last_chest = graph.add_location(NodeId.CorMineshaftLastChest, [])
    transport_to_remembrance = graph.add_location(
        NodeId.TransportToRemembrance,
        [
            stat_bonus(
                72, CheckLocation.TransportToRemembrance, [hb, locationType.TTR]
            ),
        ],
    )

    graph.register_superboss(data_demyx)

    graph.add_edge(borough, borough_chests)
    graph.add_edge(postern, postern_chests)
    graph.add_edge(corridors, corridors_chests)
    graph.add_edge(ansems_study, ansems_study_chests)
    graph.add_edge(ukulele_charm, ukulele_charm_chests)
    graph.add_edge(restoration_site, restoration_site_chests)
    graph.add_edge(crystal_fissure, crystal_fissure_chests)
    graph.add_edge(gull_wing, gull_wing_chests)
    graph.add_edge(heartless_manufactory, heartless_manufactory_chests)

    graph.add_edge(cor_depths, cor_depths_chests)
    graph.add_edge(cor_mineshaft_pre_fight_1, cor_mineshaft_pre_fight_1_chests)
    graph.add_edge(cor_depths_post_fight_1, cor_depths_post_fight_1_chests)
    graph.add_edge(cor_mining_area, cor_mining_area_chests)
    graph.add_edge(cor_mineshaft_pre_fight_2, cor_mineshaft_pre_fight_2_chests)
    graph.add_edge(cor_engine_chamber, cor_engine_chamber_chests)
    graph.add_edge(cor_mineshaft_post_fight_2, cor_mineshaft_post_fight_2_chests)
    graph.add_edge(cor_mineshaft_last_chest, cor_mineshaft_last_chest_chests)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, marketplace_map_popup)
        graph.add_edge(marketplace_map_popup, borough)
        graph.add_edge(borough, merlins_house_popup)
        graph.add_edge(merlins_house_popup, bailey, RequirementEdge(battle=True))
        graph.add_edge(bailey, baseball_charm_popup)
        graph.add_edge(baseball_charm_popup, postern)
        graph.add_edge(postern, corridors)
        graph.add_edge(corridors, dtd_popup)
        graph.add_edge(corridors, ansems_study)
        graph.add_edge(dtd_popup, ukulele_charm)

        graph.add_edge(dtd_popup, cor_depths)
        graph.add_edge(cor_depths, cor_mineshaft_pre_fight_1)
        graph.add_edge(
            cor_mineshaft_pre_fight_1,
            cor_depths_post_fight_1,
            RequirementEdge(battle=True),
        )
        graph.add_edge(cor_depths_post_fight_1, cor_mining_area)
        graph.add_edge(cor_mining_area, cor_mineshaft_pre_fight_2)
        graph.add_edge(
            cor_mineshaft_pre_fight_2, cor_engine_chamber, RequirementEdge(battle=True)
        )
        graph.add_edge(cor_engine_chamber, cor_mineshaft_post_fight_2)
        graph.add_edge(cor_mineshaft_post_fight_2, cor_mineshaft_last_chest)
        graph.add_edge(
            cor_mineshaft_last_chest,
            transport_to_remembrance,
            RequirementEdge(battle=True),
        )

        graph.add_edge(dtd_popup, restoration_site, RequirementEdge(battle=True))
        graph.add_edge(restoration_site, demyx, RequirementEdge(battle=True))
        graph.add_edge(demyx, final_fantasy_fights, RequirementEdge(battle=True))
        graph.add_edge(final_fantasy_fights, crystal_fissure)
        graph.add_edge(
            crystal_fissure, thousand_heartless, RequirementEdge(battle=True)
        )
        graph.add_edge(thousand_heartless, gull_wing)
        graph.add_edge(thousand_heartless, heartless_manufactory)
        graph.add_edge(thousand_heartless, sephiroth, RequirementEdge(battle=True))
        graph.add_edge(thousand_heartless, mushroom_13)
        graph.add_edge(thousand_heartless, data_demyx, RequirementEdge(battle=True))
        graph.register_first_boss(baseball_charm_popup)
        graph.register_last_story_boss(thousand_heartless)
        graph.register_superboss(sephiroth)
    else:
        graph.add_edge(START_NODE, borough)
        graph.add_edge(borough, postern)
        graph.add_edge(postern, corridors)
        graph.add_edge(corridors, dtd_popup)
        graph.add_edge(corridors, ansems_study)
        graph.add_edge(dtd_popup, ukulele_charm)
        graph.add_edge(dtd_popup, restoration_site, RequirementEdge(battle=True))
        graph.add_edge(restoration_site, demyx, RequirementEdge(battle=True))
        graph.add_edge(demyx, final_fantasy_fights, RequirementEdge(battle=True))
        graph.add_edge(final_fantasy_fights, crystal_fissure)
        graph.add_edge(
            crystal_fissure, thousand_heartless, RequirementEdge(battle=True)
        )
        graph.add_edge(thousand_heartless, gull_wing)
        graph.add_edge(thousand_heartless, heartless_manufactory)
        graph.add_edge(thousand_heartless, cor_depths)
        graph.add_edge(cor_depths, cor_mineshaft_pre_fight_1)
        graph.add_edge(
            cor_mineshaft_pre_fight_1,
            cor_depths_post_fight_1,
            RequirementEdge(battle=True),
        )
        graph.add_edge(cor_depths_post_fight_1, cor_mining_area)
        graph.add_edge(cor_mining_area, cor_mineshaft_pre_fight_2)
        graph.add_edge(
            cor_mineshaft_pre_fight_2, cor_engine_chamber, RequirementEdge(battle=True)
        )
        graph.add_edge(cor_engine_chamber, cor_mineshaft_post_fight_2)
        graph.add_edge(cor_mineshaft_post_fight_2, cor_mineshaft_last_chest)
        graph.add_edge(
            cor_mineshaft_last_chest,
            transport_to_remembrance,
            RequirementEdge(battle=True),
        )
        graph.add_edge(transport_to_remembrance, mushroom_13)
        graph.add_edge(transport_to_remembrance, sephiroth)
        graph.add_edge(sephiroth, marketplace_map_popup)
        graph.add_edge(marketplace_map_popup, merlins_house_popup)
        graph.add_edge(merlins_house_popup, bailey, RequirementEdge(battle=True))
        graph.add_edge(bailey, baseball_charm_popup)
        graph.add_edge(baseball_charm_popup, data_demyx, RequirementEdge(battle=True))
        graph.register_first_boss(thousand_heartless)
        graph.register_last_story_boss(bailey)



from ..List.configDict import locationType
from . import encoding


class ProgressionPoints:
    def __init__(self):
        self.gen_to_tracker_location_map = {
            locationType.Level: "Levels",
            locationType.FormLevel: "Drives",
            locationType.STT: "SimulatedTwilightTown",
            locationType.TT: "TwilightTown",
            locationType.HB: "HollowBastion",
            locationType.BC: "BeastsCastle",
            locationType.OC: "OlympusColiseum",
            locationType.Agrabah: "Agrabah",
            locationType.LoD: "LandofDragons",
            locationType.HUNDREDAW: "HundredAcreWood",
            locationType.PL: "PrideLands",
            locationType.Atlantica: "Atlantica",
            locationType.DC: "DisneyCastle",
            locationType.HT: "HalloweenTown",
            locationType.PR: "PortRoyal",
            locationType.SP: "SpaceParanoids",
            locationType.TWTNW: "TWTNW",
            locationType.CoR: "CavernofRemembrance",
        }
        self.initialize_points()
        self.initialize_point_thresholds()

    def get_world_options(self):
        return list(self.gen_to_tracker_location_map.keys())

    def get_num_cp_for_world(self, world: locationType):
        return len(self.points[world])

    def get_cp_for_world(self, world: locationType, index: int):
        return self.points[world][index]

    def get_cp_label_for_world(self, world: locationType, index: int):
        human_readable_labels = {
            "Levels": ["Level 10", "Level 20", "Level 30", "Level 40", "Level 50"],
            "Drives": [
                "Drive Level 2",
                "Drive Level 3",
                "Drive Level 4",
                "Drive Level 5",
                "Drive Level 6",
                "Drive Level 7",
            ],
            "SimulatedTwilightTown": [
                "STT Enter",
                "Olette Munny Pouch",
                "Twilight Thorn",
                "Axel 1",
                "Setzer",
                "Mansion Computer Room",
                "Axel 2",
                "Data Roxas",
            ],
            "TwilightTown": [
                "Twilight Town Enter",
                "Station Dusk Fight",
                "Mysterious Tower",
                "Sandlot Fight",
                "Mansion Fight",
                "Betwixt and Between",
                "Data Axel",
            ],
            "HollowBastion": [
                "Hollow Bastion Enter",
                "Bailey Fight",
                "Ansem's Study Computer",
                "Corridors Fight",
                "Dancers Fight",
                "Demyx",
                "Final Fantasy Fights",
                "1000 Heartless",
                "Sephiroth",
                "Data Demyx",
            ],
            "LandofDragons": [
                "Land of Dragons Enter",
                "Mission 3",
                "Mountain Trail",
                "Cave Fight",
                "Summit Fight",
                "Shan Yu",
                "Antechamber Nobodies",
                "Stormrider",
                "Data Xigbar",
            ],
            "BeastsCastle": [
                "Beast's Castle Enter",
                "Thresholder",
                "Beast's Room",
                "Dark Thorn",
                "Ballroom Dragoons Fight",
                "Xaldin",
                "Data Xaldin",
            ],
            "OlympusColiseum": [
                "Olympus Coliseum Enter",
                "Cerberus",
                "Phil's Urns",
                "Demyx",
                "Pete",
                "Hydra",
                "Auron Statue",
                "Hades",
                "AS Zexion",
                "Data Zexion",
            ],
            "DisneyCastle": [
                "Disney Castle Enter",
                "Minnie Escort",
                "Old Pete",
                "Timeless River Windows",
                "Boat Pete",
                "Pete",
                "AS Marluxia",
                "Data Marluxia",
                "Lingering Will",
            ],
            "PortRoyal": [
                "Port Royal Enter",
                "Town Fight",
                "1 Minute Fight",
                "Boat Medallion",
                "Boat Barrels",
                "Barbossa",
                "Grim Reaper 1",
                "First Medallion Gambler",
                "Grim Reaper 2",
                "Data Luxord",
            ],
            "Agrabah": [
                "Agrabah Enter",
                "Abu Minigame",
                "Chasm of Challenges",
                "Treasure Room",
                "Twin Lords",
                "Carpet Magic",
                "Genie Jafar",
                "AS Lexaeus",
                "Data Lexaeus",
            ],
            "HalloweenTown": [
                "Halloween Town Enter",
                "Candy Cane Lane Fight",
                "Prison Keeper",
                "Oogie Boogie",
                "Lock, Shock, Barrel",
                "Presents Minigame",
                "Experiment",
                "AS Vexen",
                "Data Vexen",
            ],
            "PrideLands": [
                "Pride Lands Enter",
                "Talking to Simba",
                "Hyenas 1",
                "Scar",
                "Hyenas 2",
                "Groundshaker",
                "Data Saix",
            ],
            "SpaceParanoids": [
                "Space Paranoids Enter",
                "Screens",
                "Hostile Program",
                "Solar Sailer Fight",
                "MCP",
                "AS Larxene",
                "Data Larxene",
            ],
            "HundredAcreWood": [
                "Pooh's Howse Enter",
                "Piglet's Howse Complete",
                "Rabbit's Howse Complete",
                "Kanga's Howse Complete",
                "Spooky Cave Complete",
                "Starry Hill Complete",
            ],
            "Atlantica": [
                "Atlantica Enter",
                "Ursula's Revenge",
                "A New Day is Dawning",
            ],
            "TWTNW": [
                "TWTNW Enter",
                "Roxas",
                "Xigbar",
                "Luxord",
                "Saix",
                "Xemnas 1",
                "Data Xemnas",
            ],
            "CavernofRemembrance": [
                "CoR Enter",
                "First Fight",
                "Steam Valves",
                "Second Fight",
                "Transport to Remembrance",
            ],
        }
        return human_readable_labels[self.gen_to_tracker_location_map[world]][index]

    def set_cp_for_world(self, world: locationType, index: int, value: int):
        self.points[world][index] = value

    def initialize_point_thresholds(self):
        self.point_thresholds = [
            3,
            3,
            3,
            4,
            4,
            4,
            5,
            5,
            5,
            6,
            6,
            6,
            7,
            7,
            7,
            8,
            8,
            8,
        ]  # max 18 (17 worlds + creations)

    def initialize_points(self):
        self.points = {
            locationType.Level: [0, 0, 1, 1, 2],
            locationType.FormLevel: [0, 0, 0, 1, 1, 1],
            locationType.STT: [0, 1, 2, 1, 1, 1, 1, 6],
            locationType.TT: [0, 1, 1, 2, 2, 2, 6],
            locationType.HB: [0, 1, 0, 2, 1, 2, 2, 2, 5, 6],
            locationType.BC: [0, 2, 1, 2, 1, 3, 6],
            locationType.OC: [0, 2, 0, 2, 0, 2, 2, 3, 3, 6],
            locationType.Agrabah: [0, 1, 1, 1, 2, 1, 3, 4, 6],
            locationType.LoD: [0, 1, 1, 1, 1, 2, 0, 3, 6],
            locationType.HUNDREDAW: [0, 1, 1, 1, 2, 4],
            locationType.PL: [0, 1, 1, 2, 1, 3, 6],
            locationType.Atlantica: [0, 5, 3],
            locationType.DC: [0, 1, 1, 2, 1, 2, 5, 6, 7],
            locationType.HT: [0, 1, 2, 2, 1, 1, 3, 3, 6],
            locationType.PR: [0, 1, 1, 0, 1, 2, 1, 1, 3, 6],
            locationType.SP: [0, 1, 2, 2, 2, 4, 6],
            locationType.TWTNW: [0, 2, 3, 2, 2, 2, 7],
            locationType.CoR: [0, 4, 0, 4, 7],
        }

    def get_compressed(self) -> str:
        keys = list(self.points.keys())

        threshold_characters = ""
        for p in self.point_thresholds:
            threshold_characters += str(p)

        points_characters = ""
        for k in keys:
            for i in range(len(self.points[k])):
                points_characters += str(self.points[k][i])

        threshold_result = encoding.v2r(int(threshold_characters))
        points_result = encoding.v2r(int(points_characters))
        result = threshold_result + "+" + points_result

        return result

    def set_uncompressed(self, compressed_string: str):
        keys = list(self.points.keys())

        split_strings = compressed_string.split("+")

        threshold_count = len(self.point_thresholds)
        decoded_threshold = str(encoding.r2v(split_strings[0]))
        threshold_characters = self._prepend_zeroes(
            decoded_threshold, expected_length=threshold_count
        )
        for i in range(threshold_count):
            self.set_point_threshold(i, int(threshold_characters[i]))

        points_expected_length = 0
        for k in keys:
            points_expected_length = points_expected_length + len(self.points[k])
        decoded_points = str(encoding.r2v(split_strings[1]))
        points_characters = self._prepend_zeroes(
            decoded_points, expected_length=points_expected_length
        )

        current_index = 0
        for k in keys:
            location_points: list[int] = self.points[k]
            for i in range(len(location_points)):
                location_points[i] = int(points_characters[current_index])
                current_index += 1

        return True

    def get_point_threshold(self, index):
        return self.point_thresholds[index]

    def set_point_threshold(self, index, value):
        self.point_thresholds[index] = value

    def get_hint_thresholds(self, world_count):
        output = [
            self.point_thresholds[x]
            for x in range(len(self.point_thresholds))
            if x < world_count
        ]
        return output

    def get_points_json(self):
        output = {}
        for l in locationType:
            if l in self.points:
                output[self.gen_to_tracker_location_map[l]] = self.points[l]
        return output

    def settings_spoiler_json(self) -> dict[str, list[int]]:
        output = {"Hint Thresholds": self.point_thresholds}
        for location in locationType:
            if location in self.points:
                output[locationType(location).value] = self.points[location]
        return output

    @staticmethod
    def _prepend_zeroes(value: str, expected_length: int) -> str:
        difference = expected_length - len(value)
        return ("0" * difference) + value
